#!/usr/bin/python

import numpy as np
from itertools import takewhile
from scipy.constants import h, c, e
import logging


# Returns a list of type "dtype".
def read_stream(ole, key, dtype):
    if ole.exists(key):
        stream = ole.openstream(key)
        return np.frombuffer(stream.getvalue(), dtype).tolist()
    else:
        logging.error("Stream %s does not exist in ole file", key)


def read_imageinfo_as_int(ole, key_part, ref_data=False):
    stream_key = f"ImageInfo/{key_part}"
    if ref_data:
        stream_key = "ReferenceData/" + stream_key
    integer_tuple = read_stream(ole, stream_key, "i")
    if integer_tuple is not None:
        return integer_tuple[0]


def extract_single_image(ole, numimage, numrows, numcols):
    # Read the images - They are stored in the txrm as ImageData1 ...
    # Each folder contains 100 images 1-100, 101-200
    img_key = f"ImageData{int(np.ceil(numimage / 100.0))}/Image{numimage}"
    if ole.exists(img_key):
        img_stream_bytes = ole.openstream(img_key).getvalue()
        img_stream_length = len(img_stream_bytes)
        img_size = numrows * numcols
        if (img_stream_length == (img_size * 2)):
            imgdata = np.frombuffer(img_stream_bytes, dtype=np.uint16)
        elif (img_stream_length == (img_size * 4)):
            imgdata = np.frombuffer(img_stream_bytes, dtype=np.float32)
        else:
            logging.error("Unexpected data type with %g bytes per pixel", (img_stream_length / img_size))
            raise TypeError("Reference is stored as unexpected type. Expecting uint16 or float32.")
        imgdata.shape = (numrows, numcols)
        return imgdata
    else:
        return np.zeros([])


def extract_image_dims(ole):
    return [read_imageinfo_as_int(ole, "ImageHeight"),
            read_imageinfo_as_int(ole, "ImageWidth")]


def extract_ref_dims(ole):
    if ole.exists("ReferenceData/ImageInfo/ImageHeight"):
        return [read_imageinfo_as_int(ole, "ImageHeight", ref_data=True),
                read_imageinfo_as_int(ole, "ImageWidth", ref_data=True)]
    return extract_image_dims(ole)


def extract_number_of_images(ole):
    return read_imageinfo_as_int(ole, "NoOfImages")


def extract_tilt_angles(ole):
    return read_stream(ole, "ImageInfo/Angles", "f")


def extract_x_coords(ole):
    return np.asarray(read_stream(ole, "ImageInfo/XPosition", "f"))


def extract_y_coords(ole):
    return np.asarray(read_stream(ole, "ImageInfo/YPosition", "f"))


def extract_exposure_time(ole):
    if ole.exists('ImageInfo/ExpTimes'):
        # Returns the exposure of the image at the closest angle to 0 degrees:
        exposures = read_stream(ole, "ImageInfo/ExpTimes", "f")
        absolute_angles = np.array([abs(angle) for angle in extract_tilt_angles(ole)])
        min_angle = absolute_angles.argmin()
        return exposures[min_angle]
    elif ole.exists("ImageInfo/ExpTime"):
        return read_stream(ole, "ImageInfo/ExpTime", dtype="f")[0]
    else:
        raise Exception("No exposure time available in ole file.")


def extract_multiple_exposure_times(ole):
    if ole.exists('ImageInfo/ExpTimes'):
        # Returns the exposure of the image at the closest angle to 0 degrees:
        exposures = read_stream(ole, "ImageInfo/ExpTimes", "f")
        return exposures
    elif ole.exists("ImageInfo/ExpTime"):
        return read_stream(ole, "ImageInfo/ExpTime", dtype="f")
    else:
        raise Exception("No exposure time available in ole file.")


def extract_pixel_size(ole):
    return read_stream(ole, 'ImageInfo/PixelSize', "f")[0]


def extract_all_images(ole):
    num_rows, num_columns = extract_image_dims(ole)
    if ole.exists("ImageInfo/ExpTimes"):
        images_taken = read_imageinfo_as_int(ole, "ImagesTaken")
    elif ole.exists("ImageInfo/ExpTime"):
        images_taken = 1
    if (num_rows * num_columns > 0):
        # Iterates through images until the number of images taken
        # lambda check has been left in in case stream is wrong
        images = (extract_single_image(ole, i, num_rows, num_columns) for i in range(1, images_taken + 1))
        return list(takewhile(lambda image: image.size > 1, images))
    else:
        raise AttributeError("No image dimensions found")


def extract_first_image(ole):
    return extract_all_images(ole)[0]


def extract_xray_magnification(ole):
    return read_stream(ole, "ImageInfo/XrayMagnification", "f")[0]


def extract_energy(ole):
    energies = read_stream(ole, "ImageInfo/Energy", "f")
    return np.mean(energies)


def extract_wavelength(ole):
    return h * c / (extract_energy(ole) * e)


def create_reference_mosaic(ole, refdata, image_rows, image_columns, mosaic_rows, mosaic_columns):
    ref_num_rows = convert_to_int(image_rows / mosaic_rows)
    ref_num_columns = convert_to_int(image_columns / mosaic_columns)
    refdata.shape = (ref_num_rows, ref_num_columns)
    return np.tile(refdata, (mosaic_rows, mosaic_columns))


def rescale_ref_exposure(ole, refdata):
    # TODO: Improve this in line with ALBA's methodolgy
    # Normalises the reference exposure
    # Assumes roughly linear response, which is a reasonable estimation
    # because exposure times are unlikely to be significantly different)
    # (if it is a tomo, it does this relative to the 0 degree image, not on a per-image basis)
    ref_exp = read_stream(ole, "ReferenceData/ExpTime", dtype="f")[0]
    im_exp = extract_exposure_time(ole)
    ref_exp_rescale = im_exp / ref_exp
    if ref_exp_rescale == 1:
        return refdata
    return refdata * ref_exp_rescale


def extract_reference_image(ole):
    # Read the reference image.
    # Reference info is stored under 'ReferenceData/...'
    num_rows, num_columns = extract_ref_dims(ole)
    ref_stream_bytes = ole.openstream("ReferenceData/Image").getvalue()
    img_size = num_rows * num_columns

    #  In XMController 13+ the reference file is not longer always a float. Additionally, there
    #  are multiple methods to apply a reference now, so this has been kept general, rather than
    #  being dependent on the file version or software version (software version is new metadata
    #  introduced in XMController 13).
    
    # Version 10 style mosaic:
    is_mosaic = ole.exists("ImageInfo/MosiacMode") and read_imageinfo_as_int(ole, "MosiacMode") == 1
    # This MosiacMode has been removed in v13 but is kept in for backward compatibility:
    # ImageInfo/MosiacMode (genuinely how it's spelled in the file) == 1 if it is a mosaic, 0 if not.
    if is_mosaic:
        mosaic_rows = read_imageinfo_as_int(ole, "MosiacRows")
        mosaic_cols = read_imageinfo_as_int(ole, "MosiacColumns")
        mosaic_size_multiplier = mosaic_rows * mosaic_cols
    else:
        mosaic_size_multiplier = 1
    ref_stream_length = len(ref_stream_bytes)
    mosaic_stream_length = ref_stream_length * mosaic_size_multiplier
    if mosaic_stream_length == img_size * 2:
        ref_dtype = np.uint16
    elif mosaic_stream_length == img_size * 4:
        ref_dtype = np.float32
    else:
        logging.error("Unexpected data type with %g bytes per pixel", mosaic_stream_length / img_size)
        raise TypeError("Reference is stored as unexpected type. Expecting uint16 or float32.")

    refdata = np.frombuffer(ref_stream_bytes, dtype=ref_dtype)
    if is_mosaic:
        refdata = create_reference_mosaic(ole, refdata, num_rows, num_columns, mosaic_rows, mosaic_cols)
    refdata.shape = (num_rows, num_columns)

    return rescale_ref_exposure(ole, refdata)


def convert_to_int(value):
    if int(value) == float(value):
        return int(value)
    raise ValueError(f"Value '{value}' cannot be converted to an integer")



version_info = (1, 0, 3)
__version__ = '.'.join(str(c) for c in version_info)
__author__ = "Thomas Fish"


def convert_and_save(input_path, custom_reference=None, output=None, ignore_reference=False, logging_level="info"):
    """
    Arguments:
        input_path (path, can be a file or directory (a directory will batch convert all valid contents))
        custom_reference=None (file path, not valid for batch processing)
        output=None (path, for batch processing output must be a directory otherwise outputs will be saved in the same directory as each input file found)
        ignore_reference=False (boolean)
        logging_level="info" (string: debug, info, warning, error, critical)

    Saves as .ome.tiff to output
    """
    from .run import run
    run(input_path, custom_reference, output, ignore_reference, logging_level)


def convert(input_file, custom_reference=None, ignore_reference=False, logging_level="info"):
    """
    Arguments:
        input_file (file path)
        custom_reference=None (file path)
        ignore_reference=False (boolean)
        logging_level="info" (string: debug, info, warning, error, critical)

    Outputs:
        image_list (list of numpy arrays)
        metadata (OME metadata as omexml-dls object)
    """
    from .run import _convert_file
    converter = _convert_file(input_file, custom_reference, ignore_reference, logging_level)
    return converter.get_image_and_metadata()


def save(output_path, image_list, metadata=None):
    """
    Arguments:
        output_path (file path)
        image_list (list of numpy arrays)
        metadata=None (OME metadata as omexml-dls object)

    Saves as .ome.tiff to output
    """
    from .txrm_to_image import manual_save
    manual_save(output_path, image_list, metadata)



import unittest
from unittest.mock import patch, MagicMock, call

from pathlib import Path
from tempfile import TemporaryDirectory
from random import randint, sample

from txrm2tiff.run import run, _batch_convert_files, _convert_and_save, _define_output_suffix, TxrmToImage

class TestRun(unittest.TestCase):

    def test_define_output_suffix(self):
        txrm_output = _define_output_suffix(Path("file.txrm"))
        self.assertEqual("file.ome.tiff", str(txrm_output))
        xrm_output = _define_output_suffix(Path("file.xrm"))
        self.assertEqual("file.ome.tif", str(xrm_output))
        txrm_output2 = _define_output_suffix(Path("file.extension"), ".txrm")
        self.assertEqual("file.ome.tiff", str(txrm_output2))
        xrm_output2 = _define_output_suffix(Path("file.extension"), ".xrm")
        self.assertEqual("file.ome.tif", str(xrm_output2))
        with self.assertRaises(NameError):
            _define_output_suffix(Path("file.bad_extension"), None)

    def test_define_output_suffix(self):
        txrm_output = _define_output_suffix(Path("file.txrm"))
        xrm_output = _define_output_suffix(Path("file.xrm"))
        self.assertEqual("file.ome.tiff", str(txrm_output))
        self.assertEqual("file.ome.tif", str(xrm_output))

    @patch('txrm2tiff.run.file_can_be_opened', MagicMock(return_value=True))
    @patch('txrm2tiff.run.ole_file_works', MagicMock(return_value=True))
    @patch.object(TxrmToImage, 'convert')
    @patch.object(TxrmToImage, 'save')
    def test_convert_and_save(self, mocked_save, mocked_convert):
        input_filepath = Path("test_file.txrm")

        _convert_and_save(input_filepath, None, None, False)

        mocked_convert.assert_called_with(input_filepath, None, False)
        mocked_save.assert_called_with(_define_output_suffix(input_filepath))

    @patch('pathlib.Path.mkdir', MagicMock())
    @patch('txrm2tiff.run.file_can_be_opened', MagicMock(return_value=True))
    @patch('txrm2tiff.run.ole_file_works', MagicMock(return_value=True))
    @patch.object(TxrmToImage, 'convert')
    @patch.object(TxrmToImage, 'save')
    def test_convert_and_save_with_str_output(self, mocked_save, mocked_convert):
        input_filepath = Path("test_file.txrm")
        output_str = "./output/file.extension"
        _convert_and_save(input_filepath, output_str, None, False)

        mocked_convert.assert_called_with(input_filepath, None, False)
        mocked_save.assert_called_with(Path(output_str).with_suffix('.ome.tiff'))

    @patch('pathlib.Path.mkdir', MagicMock())
    @patch('txrm2tiff.run.file_can_be_opened', MagicMock(return_value=True))
    @patch('txrm2tiff.run.ole_file_works', MagicMock(return_value=True))
    @patch.object(TxrmToImage, 'convert')
    @patch.object(TxrmToImage, 'save')
    def test_convert_and_save_with_dir_Path_output(self, mocked_save, mocked_convert):
        input_filepath = Path("test_file.txrm")
        output_filepath = Path("./output/")

        _convert_and_save(input_filepath, output_filepath, None, False)

        mocked_convert.assert_called_with(input_filepath, None, False)
        mocked_save.assert_called_with(_define_output_suffix(output_filepath / input_filepath.name))

    @patch('txrm2tiff.run.file_can_be_opened', MagicMock(return_value=True))
    @patch('txrm2tiff.run.ole_file_works', MagicMock(return_value=True))
    @patch.object(TxrmToImage, 'convert')
    @patch.object(TxrmToImage, 'save')
    def test_convert_and_save_with_invalid_output(self, mocked_save, mocked_convert):
        input_filepath = Path("test_file.txrm")

        _convert_and_save(input_filepath, 12345, None, False)

        mocked_convert.assert_called_with(input_filepath, None, False)
        mocked_save.assert_called_with(_define_output_suffix(input_filepath))

    def test_batch_convert_files_basic(self):
        with patch('txrm2tiff.run._convert_and_save', MagicMock()) as mocked_convert:
            with TemporaryDirectory(dir=".") as tmpdir:
                tmppath = Path(tmpdir)
                num_files = randint(5, 10)
                fake_file_list = []
                for i in sample(range(0, 99999), num_files):
                    fake_file = (tmppath / f"{i}.txrm")
                    fake_file.touch()
                    fake_file_list.append(fake_file)
                _batch_convert_files(tmppath, None, False)
            call_list = []
            for fake_file in fake_file_list:
                output_path = _define_output_suffix(fake_file)
                call_list.append(call(fake_file, output_path, None, False))
            mocked_convert.assert_has_calls(call_list, any_order=True)

    def test_batch_convert_files_with_output_and_deep_dir(self):
        with patch('txrm2tiff.run._convert_and_save', MagicMock()) as mocked_convert:
            with TemporaryDirectory(dir=".") as tmp_in:
                with TemporaryDirectory(dir=".") as tmp_out:
                    tmp_in_path = Path(tmp_in)
                    tmp_out_path = Path(tmp_out)
                    tmp_in_path_deep = tmp_in_path / "deep" / "dirs"
                    tmp_out_deep = tmp_out_path / "deep" / "dirs"
                    tmp_in_path_deep.mkdir(parents=True)
                    num_files = randint(5, 10)
                    fake_file_list = []
                    for i in sample(range(0, 99999), num_files):
                        fake_file = (tmp_in_path_deep / f"{i}.txrm")
                        fake_file.touch()
                        fake_file_list.append(fake_file)
                    _batch_convert_files(tmp_in_path, tmp_out, False)
                    self.assertTrue(tmp_out_deep.exists())
            call_list = []
            for fake_file in fake_file_list:
                output_path = _define_output_suffix(fake_file)
                call_list.append(call(fake_file, tmp_out_deep / output_path.name, None, False))
            mocked_convert.assert_has_calls(call_list, any_order=True)

    def test_run_with_file(self):
        with patch('txrm2tiff.run._convert_and_save', MagicMock()) as mocked_convert:
            with TemporaryDirectory(dir=".") as tmp_in:
                tmp_in_filepath = Path(tmp_in) / f"{randint(0,9999)}.xrm"
                tmp_in_filepath.touch()
                run(tmp_in_filepath)
            mocked_convert.assert_called_with(tmp_in_filepath, None, None, False)
   
    def test_run_with_dir(self, ):
        with patch('txrm2tiff.run._batch_convert_files', MagicMock()) as mocked_batch_convert:
            with TemporaryDirectory(dir=".") as tmp_in:
                tmp_in_path = Path(tmp_in)
                run(tmp_in_path)
            mocked_batch_convert.assert_called_with(tmp_in_path, None, False)
